from abc import ABC, abstractmethod
import importlib
import inspect
import sys, os
import types

core_lib = os.path.dirname(os.path.realpath(__file__)) + '/../Workers'
sys.path.append("/Users/alessioricco/Documents/GitHub/GPTWorkers")
print(sys.path)


def add_module_functions_to_dict(module_name,alias):
    # core.hello()
    try:
        module=importlib.import_module(module_name)
    except ModuleNotFoundError:
        print(f"The module {module_name} was not found.")
        return
    
    FUNCTION_DICT[alias] = {}

    for(name,func) in inspect.getmembers(module,inspect.getmodule):
        if callable(func):
            try:
                args = inspect.getfullargspec(func).args
            except:
                args = []
            FUNCTION_DICT[alias][name]=FunctionDef(name,args,curry(func))



# Now you can add the functions of a module to the FUNCTION_DICT like this:
# add_module_functions_to_dict('aleph')


class Evaluable(ABC):
    """
    Abstract base class for all evaluable entities in the maryChain language.

    This class defines the interface for all classes that can be evaluated in the maryChain language. 
    It is an abstract base class (ABC), which means that it cannot be instantiated directly. 
    Instead, other classes should inherit from this class and implement the abstract methods defined here.

    Each evaluable entity in maryChain is represented by an instance of a class that inherits from Evaluable.
    This class has an abstract method `evaluate`, which should be implemented by all subclasses to define their evaluation semantics.

    The evaluate method of a class representing an evaluable entity in maryChain takes keyword arguments (`kwargs`) 
    which could be used to provide context necessary for evaluation, such as a dictionary of variable names and their values. 
    """

    @abstractmethod
    def evaluate(self, **kwargs):
        """
        Abstract method for evaluating an entity.

        This method is meant to be overridden by each concrete subclass of Evaluable. The overriding method 
        should implement the specific evaluation semantics of the entity it represents.

        Parameters:
        **kwargs: Arbitrary keyword arguments that could be used to provide context necessary for evaluation. 

        Returns:
        The result of evaluating the entity.
        """
        pass


class Expression(Evaluable):
    """
    Abstract base class for all expressions in the maryChain language.
    Inherits from the Evaluable abstract base class.
    """
    pass

class Program(Evaluable):
    def __init__(self, definitions, imports, expression):
        self.definitions = definitions
        self.imports = imports
        self.expression = expression

    def evaluate(self):
        for definition in self.definitions:
            definition.evaluate()

        for import_stmt in self.imports:
            import_stmt.evaluate()

        if self.expression:
            return self.expression.evaluate()

class Import(Expression):
    def __init__(self, module_parts, alias):
        # self.module_name = '.'.join(module_parts)
        self.module_name = module_parts
        self.alias = alias

    def evaluate(self):
        add_module_functions_to_dict(self.module_name, self.alias)



# --------------------------------------------------------------
#  external functions
def print_func(x):
    if isinstance(x, Lazy):
        x = x.evaluate()
    print(x)
    return x

def add_func(x, y):
    if isinstance(x, Lazy):
        x = x.evaluate()
    if isinstance(y, Lazy):
        y = y.evaluate()
    return x + y

def eval_func(x):
    return x.func()

def curry(func):
    """
    This function transforms a given function into a curried function.
    
    Currying is a technique in functional programming where a function with multiple arguments is 
    transformed into a sequence of functions, each with a single argument. 
    For example, a function of two arguments, f(x, y), produces a function of one argument, g(y) = f(x, y), 
    which can be called with one argument to produce the result.

    Parameters:
    func (function): The function to be curried.

    Returns:
    function: The curried version of the original function.
    """
    
    def curried(*args, **kwargs):
        """
        This function represents the curried version of the original function.
        
        It tries to call the original function with the provided arguments. 
        If a TypeError occurs because one argument is missing, it returns a new function that 
        takes one argument and calls the original function with the previously provided arguments 
        and this new argument.
        """
        try:
            return func(*args, **kwargs)
        except TypeError as e:
            if 'missing 1 required positional argument' in str(e):
                return lambda x: func(*args, x, **kwargs)
            else:
                raise e

    return curried  # Return the curried function.


# def curry(func):
#     def curried(*args, **kwargs):
#         if len(args) + len(kwargs) >= func.__code__.co_argcount:
#             return func(*args, **kwargs)
#         else:
#             # Return a new FunctionDef instance
#             return FunctionDef(func.__name__, func.__code__.co_varnames[len(args):], curried)
#     return curried



# --------------------------------------------------------------
#  environment

ENVIRONMENT = {}

# --------------------------------------------------------------

class FunctionDef:
    """
    Represents a function definition in the maryChain language. Functions can either be user-defined or 
    built-in functions.
    """

    def __init__(self, name, args, body):
        """
        Initialize a function definition with a name, a list of argument names, and a body.
        
        Parameters:
        name (str): The name of the function.
        args (list): The names of the arguments of the function.
        body (expression): The body of the function which can be an expression or a Python callable 
                           (for built-in functions).
        """
        self.name = name
        self.args = args
        self.expected_args = len(args)

        # Check if the body is a function, in which case this is a built-in function
        if callable(body):
            self.is_builtin = True
            self.func = body
        else:
            self.is_builtin = False
            self.body = body

    def evaluate(self, **kwargs):
        """
        Evaluates the function definition by adding it to the global environment and returning itself.
        
        Returns:
        FunctionDef: This function definition.
        """
        if len(self.args) == 0:
            return self.body
        ENVIRONMENT[self.name] = self
        return self

    # def __call__(self, *args):
        """
        Defines how the function definition behaves when called. If the function is built-in, it simply 
        calls the function with the provided arguments. Otherwise, it updates the local environment with 
        the function arguments and evaluates the body of the function in this environment.

        If not all arguments are provided, it returns a new function that takes the remaining arguments.

        Parameters:
        args (list): The arguments to call the function with.

        Returns:
        expression: The result of evaluating the function body.
        """
        # if self.is_builtin:
        #     return self.func(*args)
        
        # if len(args) > len(self.args):
        #     raise TypeError(f"{self.name} takes {len(self.args)} arguments but {len(args)} were given")

        # # Make a copy of the global environment and update it with the function arguments
        # local_env = ENVIRONMENT.copy()
        # local_env.update(zip(self.args, args))

        # if len(args) < len(self.args):
        #     # Not all arguments provided, return a new function that takes the remaining arguments
        #     remaining_args = self.args[len(args):]
        #     return curry(FunctionDef(self.name, remaining_args, self.body).evaluate(local_env))

        # # Evaluate the body of the function in the local environment and return the result
        # return self.body.evaluate(local_env)

    def __call__(self,*args):
        if self.is_builtin:
            return curry(self.func)(*args)
        if len(args)>len(self.args):
            raise TypeError(f"{self.name} takes {len(self.args)} arguments but {len(args)} were given")
        local_env=ENVIRONMENT.copy()
        local_env.update(zip(self.args,args))
        if len(args)<len(self.args):
            remaining_args=self.args[len(args):]
            return curry(FunctionDef(self.name,remaining_args,self.body).evaluate(local_env))
        return self.body.evaluate(local_env)

class FunctionCall(Expression):
    def __init__(self, func, args):
        self.func = func
        self.args = args

    def evaluate(self, **kwargs):
        func = None
        if isinstance(self.func, Identifier):
            # Look up function by name
            name = self.func.name
            if self.func.name in ENVIRONMENT:
                func = ENVIRONMENT[self.func.name]
            else:
                # namespace
                if "." in name:  # Check if func is a qualified function name
                    module_name, func_name = name.split(".")
                    if (module_name not in FUNCTION_DICT) or (func_name not in FUNCTION_DICT[module_name]):
                        raise ValueError(f"Function {self.func} is not defined")
                    func = FUNCTION_DICT[module_name][func_name]
                else:
                    func = FUNCTION_DICT[self.func.name]
        elif isinstance(self.func, FunctionCall):
            # Evaluate inner function call to get function
            func = self.func.evaluate()
        else:
            raise ValueError(f"Function {self.func} is not defined")

        if func.is_builtin:
            # If the function is a built-in function, call it with the provided arguments
            # This will raise an exception if there are too few arguments
            evaluated_args = [arg.evaluate() for arg in self.args]
            return func(*evaluated_args)
        else:
            # If the function is a user-defined function and there are too few arguments, return a new FunctionDef for currying
            if len(self.args) < func.expected_args:
                return FunctionDef(func.name, func.args[len(self.args):], func.body)

            # If there are enough arguments, evaluate the function call as normal
            evaluated_args = [arg.evaluate() for arg in self.args]
            return func(*evaluated_args)


class LetIn(Expression):
    """
    This class represents a "let-in" expression in the maryChain language. A let-in expression
    is used to bind a value to an identifier in a certain scope. The syntax is as follows:

        let <identifier> = <expression> in <body>

    After the expression is evaluated, the result is bound to the identifier, and this binding
    is valid within the body of the let-in expression.
    """

    def __init__(self, identifier, value_expression, body):
        """
        Initializes a new instance of the LetIn class.

        Args:
            identifier: The identifier to which the value will be bound.
            value_expression: The expression that will be evaluated to determine the value
                to bind to the identifier.
            body: The body of the let-in expression, where the identifier is bound to the value.
        """
        self.identifier = identifier
        self.value_expression = value_expression
        self.body = body

    def evaluate(self):
        """
        Evaluates the let-in expression.

        This is done by first evaluating the value expression, then creating a new scope where the 
        identifier is bound to the value. The body of the let-in expression is then evaluated in this new 
        scope. Finally, the old scope is restored.

        Returns:
            The result of evaluating the body of the let-in expression.
        """
        # Evaluate the value expression and create a new scope where the identifier is bound to the value.
        value = self.value_expression.evaluate()
        new_env = ENVIRONMENT.copy()
        new_env[self.identifier] = value

        # Evaluate the body in the new scope.
        old_env = ENVIRONMENT
        ENVIRONMENT = new_env
        try:
            return self.body.evaluate()
        finally:
            ENVIRONMENT = old_env  # Restore the old environment



class BinOp(Expression):
    """
    This class represents a binary operation in the maryChain language. A binary operation is 
    an operation that takes two operands and performs a specified operation on them.

    The supported operators are:
        '+'  - addition
        '-'  - subtraction
        '*'  - multiplication
        '/'  - division
        '&&' - logical AND
        '||' - logical OR
        '->' - logical implication
        '>'  - greater than
        '<'  - less than
        '>=' - greater than or equal to
        '<=' - less than or equal to
        '==' - equality (same as)
    """

    def __init__(self, left, op, right):
        """
        Initializes a new instance of the BinOp class.

        Args:
            left: The left operand of the binary operation.
            op: The operator of the binary operation.
            right: The right operand of the binary operation.
        """
        self.left = left
        self.op = op
        self.right = right

    def evaluate(self, **kwargs):
        """
        Evaluates the binary operation and returns the result.

        This is done by first evaluating the left and right operands, then applying the operator to the results.

        Returns:
            The result of the binary operation.

        Raises:
            ValueError: If the operator is not supported.
        """
        if self.op == '+':
            return evaluate(self.left) + evaluate(self.right)
        if self.op == '-':
            return evaluate(self.left) - evaluate(self.right)
        if self.op == '*':
            return evaluate(self.left) * evaluate(self.right)
        if self.op == '/':
            return evaluate(self.left) / evaluate(self.right)
        if self.op == '&&':
            return evaluate(self.left) and evaluate(self.right)
        if self.op == '||':
            return evaluate(self.left) or evaluate(self.right) 
        if self.op == '->':
            return not evaluate(self.left) or evaluate(self.right)
        if self.op == '>':
            return evaluate(self.left) > evaluate(self.right) 
        if self.op == '<':
            return evaluate(self.left) < evaluate(self.right)    
        if self.op == '>=':
            return evaluate(self.left) >= evaluate(self.right)   
        if self.op == '<=':
            return evaluate(self.left) <= evaluate(self.right) 
        if self.op == '==':
            return evaluate(self.left) == evaluate(self.right) 
        raise ValueError(f'Unknown operator: {self.operator}')


class Number(Expression):
    """
    This class represents a numeric literal in the maryChain language. 

    A numeric literal can be an integer or a floating-point number, and this class can handle both.
    """

    def __init__(self, value):
        """
        Initializes a new instance of the Number class.

        Args:
            value: The numeric value represented by the instance. Can be either an integer or a float.

        Note:
            If the value cannot be converted to an integer, it will be converted to a float.
        """
        self.value = value
        try:
            self.value = int(value)
        except ValueError:
            self.value = float(value)

    def evaluate(self, **kwargs):
        """
        Evaluates the number, which in practice means returning the numeric value.

        Returns:
            The numeric value of the Number instance.
        """
        return self.value


class String(Expression):
    """
    This class represents a string literal in the maryChain language. 

    The `String` class inherits from the `Expression` class and provides methods for the manipulation of string values.
    """

    def __init__(self, value):
        """
        Initializes a new instance of the String class.

        Args:
            value: The string value represented by the instance.
        """
        self.value = value

    def evaluate(self, **kwargs):
        """
        Evaluates the string, which in practice means returning the string value.

        Returns:
            The string value of the String instance.
        """
        return self.value


class Boolean(Expression):
    """
    This class represents a boolean literal in the maryChain language. 

    The `Boolean` class inherits from the `Expression` class and provides methods for the manipulation of boolean values.
    """

    def __init__(self, value):
        """
        Initializes a new instance of the Boolean class.

        Args:
            value: The boolean value represented by the instance.
        """
        self.value = bool(value)

    def evaluate(self, **kwargs):
        """
        Evaluates the boolean, which in practice means returning the boolean value.

        Returns:
            The boolean value of the Boolean instance.
        """
        return self.value


class Identifier(Expression):
    """
    This class represents an Identifier in the maryChain language. 

    The `Identifier` class inherits from the `Expression` class and provides a way to interact with variables
    and function calls.
    """

    def __init__(self, name):
        """
        Initializes a new instance of the Identifier class.

        Args:
            name: The name of the variable or function this identifier refers to.
        """
        self.name = name

    def evaluate(self, **kwargs):
        """
        Evaluates the Identifier. If the identifier name is in the environment, its value is returned. 
        If the name corresponds to a function in the function dictionary, the name is returned. 
        If the name is an instance of a function call, it is evaluated.

        Returns:
            The value of the Identifier if it exists in the environment, its name if it exists in the function dictionary,
            or the result of its evaluation if it is a FunctionCall.
        """
        if self.name in ENVIRONMENT:
            return ENVIRONMENT[self.name]
        if self.name in FUNCTION_DICT:
            return self.name
        if isinstance(self.name, FunctionCall):
            return self.name.evaluate()
        return None


class UnaryOperation(Expression):
    """
    This class represents a unary operation in the maryChain language. 

    The `UnaryOperation` class inherits from the `Expression` class and encapsulates an operator and 
    an operand for operations such as unary minus and logical not.
    """

    def __init__(self, operator, operand):
        """
        Initializes a new instance of the UnaryOperation class.

        Args:
            operator: The operator of the unary operation. It could be '-' for unary minus or 'not' for logical not.
            operand: The operand on which the unary operation is to be performed.
        """
        self.operator = operator
        self.operand = operand

    def evaluate(self, **kwargs):
        """
        Evaluates the UnaryOperation. If the operator is '-', the negation of the operand is returned.
        If the operator is 'not', the logical not of the operand is returned.

        Returns:
            The result of the unary operation on the operand.

        Raises:
            ValueError: If the operator is not recognized.
        """
        if self.operator == '-':
            return -evaluate(self.operand)
        if self.operator == 'not':
            return not evaluate(self.operand)
        raise ValueError(f'Unknown operator: {self.operator}')

class While(Expression):
    """
    This class represents a while loop in the maryChain language. 

    The `While` class inherits from the `Expression` class and encapsulates a condition and a body.
    The body of the while loop is executed as long as the condition evaluates to True.
    """

    def __init__(self, condition, body):
        """
        Initializes a new instance of the While class.

        Args:
            condition: The condition of the while loop. It should be an expression that evaluates to a boolean.
            body: The body of the while loop. It's a sequence of expressions that will be executed 
                  as long as the condition is True.
        """
        self.condition = condition
        self.body = body

    def evaluate(self, **kwargs):
        """
        Evaluates the While loop. The condition is checked before each iteration. 
        If the condition is True, the body is executed, otherwise the loop is exited.

        Returns:
            The result of the last expression evaluated in the body of the loop.

        Raises:
            None
        """
        result = self.condition.evaluate()
        while result != False:
            result = self.body.evaluate()
            if self.condition.evaluate() == False:
                break
        return result

    
class IfThenElse(Expression):
    """
    This class represents a conditional "if-then-else" statement in the maryChain language.

    The `IfThenElse` class inherits from the `Expression` class and encapsulates a condition,
    an expression to evaluate if the condition is true, and another expression to evaluate if
    the condition is false.
    """

    def __init__(self, condition, true_expr, false_expr):
        """
        Initializes a new instance of the IfThenElse class.

        Args:
            condition: The condition of the if-then-else statement. It should be an expression that evaluates to a boolean.
            true_expr: The expression that will be evaluated if the condition is true.
            false_expr: The expression that will be evaluated if the condition is false.
        """
        self.condition = condition
        self.true_expr = true_expr
        self.false_expr = false_expr

    def evaluate(self, **kwargs):
        """
        Evaluates the IfThenElse statement. 

        The condition is evaluated first. If the condition is True, the true_expr is evaluated and its 
        result is returned. If the condition is False, the false_expr is evaluated and its result is returned.

        Returns:
            The result of the evaluated true_expr if the condition is true, or the result of the evaluated
            false_expr if the condition is false.

        Raises:
            None
        """
        if evaluate(self.condition):
            return evaluate(self.true_expr)
        else:
            return evaluate(self.false_expr)

        
class IfThen(Expression):
    """
    The IfThen class represents a conditional expression in a programming language. 
    It contains a condition and an expression that gets evaluated if the condition is true.
    """

    def __init__(self, condition, true_expr):
        """
        Initialize IfThen with a condition and an expression to be evaluated if the condition is true.

        :param condition: The condition of the 'if' statement.
        :param true_expr: The expression to be evaluated if the condition is true.
        """
        self.condition = condition
        self.true_expr = true_expr

    def evaluate(self, **kwargs):
        """
        Evaluate the IfThen expression. If the condition is true, evaluate and return the true expression. 
        If the condition is false, return the last evaluated result in the environment. 
        If there is no last result in the environment, return None.

        :return: Result of evaluating the true expression if the condition is true, 
                 or the last result in the environment if the condition is false.
        """
        # Evaluate the condition
        if self.condition.evaluate():
            # If the condition is true, evaluate and return the true expression
            return self.true_expr.evaluate()
        elif '%lastresult%' in ENVIRONMENT:
            # If the condition is false, return the last evaluated result in the environment
            return ENVIRONMENT['%lastresult%']
        else:
            # If there is no last result in the environment, return None
            return None  



class Lazy(Expression):
    """
    The Lazy class represents a lazy evaluation of an expression in a programming language. 
    It contains a function and its evaluated state and value.
    """

    def __init__(self, func):
        """
        Initialize Lazy with a function to be lazily evaluated.

        :param func: The function to be evaluated.
        """
        self.func = func
        self.is_evaluated = False  # Indicates if the function has been evaluated
        self.value = None  # Holds the result of the evaluation

    def evaluate(self, **kwargs):
        """
        Evaluate the Lazy expression if it hasn't been evaluated before. 
        The result is cached for future uses.

        :return: Result of evaluating the function.
        """
        # If the function hasn't been evaluated before
        if not self.is_evaluated:
            # If the function is callable
            if callable(self.func):
                # Evaluate and cache the result
                self.value = evaluate(self.func())
            else:
                # If the function is not callable, evaluate and cache the result
                self.value = evaluate(self.func)
            # Mark as evaluated
            self.is_evaluated = True

        # Return the cached value
        return self.value

    def __repr__(self):
        """
        Return a string representation of the Lazy object.
        """
        # If the function has been evaluated
        if self.is_evaluated:
            return f"<Evaluated lazy object with value: {self.value}>"
        else:
            # If the function hasn't been evaluated
            return "<Unevaluated lazy object>"


class Pipe(Expression):
    """
    The Pipe class represents a pipeline operation in a programming language.
    A pipeline operation is one where the result of one operation (left) is used as the input to another operation (right).
    """

    def __init__(self, left, right):
        """
        Initialize Pipe with two expressions: left and right.

        :param left: The left expression to be evaluated first.
        :param right: The right expression that takes the result of the left expression as an argument.
        """
        self.left = left
        self.right = right

    def evaluate(self, **kwargs):
        """
        Evaluate the Pipe expression. 
        The left expression is evaluated first and its result is fed as an argument to the right expression.

        :return: Result of evaluating the right expression with the result of the left expression.
        """
        # Evaluate the left expression and store its result in a special variable '%lastresult%'
        ENVIRONMENT['%lastresult%'] = self.left.evaluate()

        # Insert the result of the left expression as the first argument of the right expression
        self.right.args.insert(0, self.left)

        # Evaluate the right expression and return the result
        return self.right.evaluate()


class LambdaFunctionValue(Expression):
    """
    The LambdaFunctionValue class represents a lambda function in a programming language.
    A lambda function is a small anonymous function that can take any number of arguments but can have only one expression.
    """

    def __init__(self, args, body):
        """
        Initialize LambdaFunctionValue with arguments and body of the lambda function.

        :param args: The arguments that the lambda function accepts.
        :param body: The expression that forms the body of the lambda function.
        """
        self.args = args
        self.body = body

    def __call__(self, *args):
        """
        Define the behavior of the LambdaFunctionValue instance when it's 'called' like a function.
        If the number of arguments provided during the call doesn't match the lambda function's arguments,
        a TypeError is raised.

        :param args: The arguments provided during the function call.
        :return: Result of evaluating the lambda function's body with the provided arguments.
        """
        # Check if the number of arguments provided during the call matches the number of arguments the function accepts
        if len(args) != len(self.args):
            raise TypeError("wrong number of arguments")

        # Create a local environment where the arguments are bound to their values
        local_env = ENVIRONMENT.copy()
        local_env.update(zip(self.args, args))

        # Evaluate the function's body in the local environment and return the result
        return self.body.evaluate(local_env)



class LambdaFunction(Expression):
    """
    The LambdaFunction class represents a lambda function definition in a programming language. 
    This class is used to create a lambda function object (LambdaFunctionValue) when evaluated.
    """

    def __init__(self, args, body):
        """
        Initialize LambdaFunction with arguments and body of the lambda function.

        :param args: The arguments that the lambda function accepts.
        :param body: The expression that forms the body of the lambda function.
        """
        self.args = args
        self.body = body

    def evaluate(self):
        """
        Evaluate the lambda function to a LambdaFunctionValue object which can be called like a function.

        :return: A LambdaFunctionValue object that represents the lambda function.
        """
        return LambdaFunctionValue(self.args, self.body)


FUNCTION_DICT = {
    # This dictionary maps function names to FunctionDef instances that implement the function.

    # The 'out' function prints its argument and returns None. 
    # It's implemented with the 'print_func' function and curried to allow partial application.
    'out': FunctionDef('out', ['x'], curry(print_func)),

    # The 'add' function adds its two arguments. 
    # It's implemented with the 'add_func' function and curried to allow partial application.
    'add': FunctionDef('add', ['x', 'y'], curry(add_func)),

    # The 'eval' function evaluates its argument as a maryChain expression. 
    # It's implemented with the 'eval_func' function and curried to allow partial application.
    'eval': FunctionDef('eval', ['x'], curry(eval_func)),
}

# def exec(s):
#     node = None
#     return evaluate(node)

def inject_parsing(parse):
    FUNCTION_DICT['exec'] = FunctionDef('exec', ['s'], curry(lambda x: evaluate(parse(x))))
    return 

def evaluate(node):
    """
    The evaluate function is used to evaluate different types of nodes in an abstract syntax tree (AST).
    It returns the evaluated result of the node.

    :param node: A node in the AST. The node could be a string, integer, float, or an instance of Expression class.
    :return: The evaluated result of the node.
    """
    if node is None:
        return None
    
    if isinstance(node, str):
        # If the node is a string, return the string itself
        return node
    
    if isinstance(node, int):
        # If the node is an integer, return the integer itself
        return node
    
    if isinstance(node, float):
        # If the node is a float, return the float itself
        return node
    
    if isinstance(node, Evaluable):
        # If the node is an instance of Expression class, evaluate the expression and return the result
        return node.evaluate()
    
    # If the node is none of the above types, raise a ValueError
    raise ValueError(f'Unknown node type: {node}')





'''BNF
<program> ::= <defn>* <expr>?
<expression> ::= <term>
              | <term> "+" <expression>
              | <term> "-" <expression>
              | <term> "*" <expression>
              | <term> "/" <expression>
              | "(" <expression> ")"
              | <term> "==" <expression>
              | <term> "<" <expression>
              | <term> ">" <expression>
              | <term> "<=" <expression>
              | <term> ">=" <expression>
              | "if" <expression> "then" <expression> "else" <expression>
              | "while" <expression> "do" <expression>
              | "lazy" <expression>
              | "{" <expression> "}"
              | "lambda" "(" <args> ")" <expression>
              | <expression> "|" <expression>

<term> ::= <factor>
        | <string>
        | <number>
        | <boolean>
        | <function_call>

<factor> ::= <number>
           | <string>
           | "true"
           | "false"

<function_call> ::= <function_call> "(" <args> ")"
                  | <identifier> "(" <args> ")"
                  | <identifier>

<args> ::= <arg>
         | <args> "," <arg>

<arg> ::= <expression>
'''

# Importing necessary libraries
from ply import lex, yacc
from maryChain.maryChainAST import Program, BinOp, UnaryOperation, Number, String, Boolean, Identifier, LetIn, LambdaFunction
from maryChain.maryChainAST import FunctionDef, FunctionCall, While, IfThenElse, IfThen, Lazy, Pipe, Import
import maryChain.maryChainAST as AST

# Define the list of tokens that the lexer will recognize
# Define the list of tokens that the lexer will recognize.
# These tokens represent the different types of elements that can appear in the language's syntax.
# Each token corresponds to a specific type of character sequence that can appear in the language's source code.
tokens = (
    # Basic types
    'IDENTIFIER', 'NUMBER', 'STRING', 'BOOLEAN',
    
    # Arithmetic operators
    'PLUS', 'MINUS', 'TIMES', 'DIVIDE', 'EQUALS',
    
    # Parentheses and brackets
    'LPAREN', 'RPAREN', 'LBRACK', 'RBRACK', 
    
    # Comma for argument separation
    'COMMA',
    
    # Control flow keywords
    'IF', 'THEN', 'ELSE', 'WHILE', 'DO', 'TRY', 'CATCH', 'FINALLY', 'LAZY',
    
    # Logical operators
    'AND', 'OR', 'NOT', 
    
    # Other operators
    'PIPE', 'IMPLIES', 
    
    # List functions
    'LEN', 'HEAD', 'TAIL', 'EMPTY', 'FOREACH',
    
    # Type cast functions
    'BOOLEANCAST', 'STRINGCAST', 'INTEGERCAST', 'DOUBLECAST',
    
    # Boolean values and null
    'TRUE', 'FALSE', 'NULL',
    
    # Function definition keyword
    'DEF',
    
    # Comparison operators
    'EQUALITY', 'GREATER', 'LESS', 'GREATEREQUAL', 'LESSEQUAL',
    
    # Dot for method calls or attribute access
    'DOT',
    
    # Braces for block structures
    'LCBRACE', 'RCBRACE',
    
    # Function, let-in, and lambda keywords
    'FUNC', 'LET', 'IN', 'LAMBDA',

    # import libraries
    'IMPORT', 'AS', 'NAMESPACE_OP',

)


# Define a dictionary of reserved words.
# These are words that have a special meaning in the language's syntax and cannot be used as identifiers.
# The keys in this dictionary are the string representations of the reserved words,
# and the values are their corresponding token types (defined above).
reserved = {
    # 'if', 'then', and 'else' are used for if-else control structures.
    'if' : 'IF',
    'then' : 'THEN',
    'else' : 'ELSE',

    # 'while' and 'do' are used for while-do loop control structures.
    'while' : 'WHILE',
    'do' : 'DO',

    # 'try', 'catch', and 'finally' are used for exception handling.
    'try' : 'TRY',
    'catch' : 'CATCH',
    'finally' : 'FINALLY',

    # 'lazy' is used for lazy evaluation.
    'lazy' : 'LAZY',
    
    # 'func' is used to define a function.
    'func' : 'FUNC',

    # 'true' and 'false' are boolean literals.
    'true' : 'TRUE',
    'false' : 'FALSE',

    # 'let' and 'in' are used for variable declarations.
    'let' : 'LET',
    'in' : 'IN',

    # 'lambda' is used for anonymous functions.
    'lambda': 'LAMBDA',
    
    # 'import' is used for importing libraries.
    'import': 'IMPORT',
    'as': 'AS'
    # add other reserved words here
}


# Define operator precedence rules.
# These rules dictate the order in which operations are performed when evaluating expressions, 
# i.e., which operations are done first.
# Each rule is a tuple where the first element is the associativity of the operators ('left' or 'right'),
# and the remaining elements are the operator(s) at that level of precedence.
precedence = (
    # The pipe operator has the lowest precedence and is left-associative.
    ('left', 'PIPE'),

    # The logical OR operator is next in line with left associativity.
    ('left', 'OR'),

    # The logical AND operator follows with left associativity.
    ('left', 'AND'),

    # The logical NOT operator is next with left associativity.
    ('left', 'NOT'),

    # The equality and relational operators all have the same level of precedence and are left-associative.
    ('left', 'EQUALITY', 'GREATER', 'LESS', 'GREATEREQUAL', 'LESSEQUAL'),

    # The addition and subtraction operators are at the next level of precedence, also left-associative.
    ('left', 'PLUS', 'MINUS'),

    # The multiplication and division operators are next with higher precedence, left-associative.
    ('left', 'TIMES', 'DIVIDE'),

    # The unary minus operator has the highest precedence and is right-associative.
    ('right', 'UMINUS'),
)


# Define regular expressions for each token
t_DEF = r'def'
# t_PLACEHOLDER = r'_'
t_PLUS = r'\+'
t_MINUS = r'-'
t_TIMES = r'\*'
t_DIVIDE = r'/'
t_EQUALS = r'='
t_LPAREN = r'\('
t_RPAREN = r'\)'
t_LBRACK = r'\['
t_RBRACK = r'\]'
t_COMMA = r','
t_AND = r'&&'
t_OR = r'\|\|'
t_NOT = r'not'
t_PIPE = r'\|'
t_IMPLIES = r'->'
t_LEN = r'len\(\)'
t_HEAD = r'head\(\)'
t_TAIL = r'tail\(\)'
t_EMPTY = r'empty\(\)'
t_FOREACH = r'foreach\('
t_BOOLEANCAST = r'boolean\(\)'
t_STRINGCAST = r'string\(\)'
t_INTEGERCAST = r'integer\(\)'
t_DOUBLECAST = r'double\(\)'
t_DOT = r'\.'
t_EQUALITY = r'=='
t_GREATER = r'>'
t_LESS = r'<'
t_GREATEREQUAL = r'>='
t_LESSEQUAL = r'<='
t_LCBRACE = r'\{'
t_RCBRACE = r'\}'
t_ignore  = ' \t'
t_IMPORT = 'import'  # Simply match the keyword
t_AS = 'as'
t_NAMESPACE_OP = r'::'

def t_IDENTIFIER(t):
    r'[a-zA-Z_][a-zA-Z_0-9]*(\.[a-zA-Z_][a-zA-Z_0-9]*)*'
    t.type = reserved.get(t.value, 'IDENTIFIER')    # Check for reserved words
    return t

def t_NUMBER(t):
    r'\d+(\.\d+)?'
    t.value = float(t.value) if '.' in t.value else int(t.value)
    return t

def t_STRING(t):
    r'(\'([^\\\n]|(\\.))*?\'|\"([^\\\n]|(\\.))*?\")'
    t.value = t.value[1:-1]  # remove the quotes
    return t

def t_BOOLEAN(t):
    r'(true|false)'
    t.value = (t.value == 'true')  # convert to Python boolean
    return t

def t_COMMENT(t):
    r'\#.*'
    pass
    # No return value. Token discarded

def t_newline(t):
    r'\n+'
    t.lexer.lineno += len(t.value)

def t_error(t):
    line = t.lexer.lineno
    position = t.lexpos
    value = t.value[0]
    print(f"Illegal character at line {line}, position {position}: {value}")
    t.lexer.skip(1)


# --------------------------------------------------------------
# Initialize the lexer
lexer = lex.lex()

def p_program(t):
    '''program : definitions imports expression
               | imports expression
               | definitions expression
               | imports
               | definitions
               | expression'''
    """
    Handles the parsing of the whole program.
    
    The function handles the parsing of the program which consists of function definitions,
    imports, and the main expression. Each component is optional and can appear zero or more
    times (for definitions and imports) or at most once (for expression).

    Parameters:
    t: A list where each element represents a component of the syntax being parsed.

    Sets:
    t[0]: A Program object, representing the whole program.
    """

    if len(t) == 2:
        imports = []
        definitions = []
        expression = t[1]
        if isinstance(t[1],list):
            if len(t[1])> 0:
                if isinstance(t[1][0], Import):
                    imports = t[1]
                    expression = None
                elif isinstance(t[1][0], FunctionDef):
                    definitions = t[1]
                    expression = None
            
        t[0] = Program(imports, definitions, expression)

    elif len(t) == 3:
        # is definition or expresssion?
        imports = []
        definitions = []
        if isinstance(t[1],list):
            if len(t[1])> 0:
                if isinstance(t[1][0], Import):
                    imports = t[1]
                elif isinstance(t[1][0], FunctionDef):
                    definitions = t[1]

        t[0] = Program(definitions, imports, t[2])
    elif len(t) == 4:
        t[0] = Program(t[1], t[2], t[3])


def p_expression_let_in(t):
    'expression : LET IDENTIFIER EQUALS expression IN expression'
    """
    Handles the parsing of 'let in' expressions.

    The function handles the 'let' keyword syntax which is used to declare variables.
    In the syntax 'let IDENTIFIER = expression in expression', an IDENTIFIER (variable name)
    is assigned the value of an expression, and this variable is then available within the
    scope of the next expression. This resembles local variable assignment in other programming
    languages.

    Parameters:
    t: A list where each element represents a component of the syntax being parsed. The elements
       correspond to the 'let IDENTIFIER = expression in expression' structure.

    Sets:
    t[0]: A LetIn object, representing the variable assignment and scope for the 'let in'
          construct.
    """
    t[0] = LetIn(t[2], t[4], t[6])



def p_expression_binop(t):
    '''expression : expression PLUS expression
                  | expression MINUS expression
                  | expression TIMES expression
                  | expression DIVIDE expression'''
    """
    Handles the parsing of binary operations.

    The function handles the basic binary mathematical operations: addition, subtraction,
    multiplication, and division. For a given operation 'expression OPERATOR expression', it
    constructs a BinOp object with the two expressions and the operator as arguments.

    Parameters:
    t: A list where each element represents a component of the syntax being parsed. The elements
       correspond to the 'expression OPERATOR expression' structure.

    Sets:
    t[0]: A BinOp object, representing the binary operation.
    """
    t[0] = BinOp(t[1], t[2], t[3])


# --------------------------------------------------------------
#  import
def p_namespace(p):
    '''namespace : IDENTIFIER
                 | namespace DOT IDENTIFIER'''
    if len(p) == 2:
        # If namespace is a single identifier, make it a one-element list
        p[0] = [p[1]]
    else:
        # If namespace is another namespace followed by an identifier, 
        # append the identifier to the list of identifiers in the namespace
        p[0] = p[1] + [p[3]]

def p_import(t):
    'import : IMPORT IDENTIFIER AS IDENTIFIER'
    print("Import statement detected: ", t[2], t[4])
    t[0] = Import(t[2], t[4])

def p_imports(t):
    '''
    imports : import
            | imports import
    '''
    # if there's only one import statement
    if len(t) == 2:
        t[0] = [t[1]]
    # if there's more than one import statement
    else:
        t[1].append(t[2])
        t[0] = t[1]


# Rule for unary minus operation
def p_expression_uminus(t):
    'expression : MINUS expression %prec UMINUS'
    t[0] = UnaryOperation('-', t[2])  # Represent unary minus as a UnaryOperation.

def p_expression_paren(t):
    'expression : LPAREN expression RPAREN'
    t[0] = t[2]  # No change here, parentheses do not change the AST.

def p_term_factor(t):
    'term : factor'
    t[0] = t[1]  # No change here, we are just passing up the node.

def p_factor_string(t):
    'factor : STRING'
    t[0] = String(t[1])  # Wrap the string in a String node.

def p_factor_num(t):
    'factor : NUMBER'
    t[0] = Number(t[1])  # This will try to cast to int first, then float

def p_factor_boolean(t):
    '''factor : TRUE
              | FALSE'''
    t[0] = Boolean(t[1] == 'true')  # This will try to cast to boolean

# --------------------------------------------------------------
#  while

def p_expression_while(t):
    'expression : WHILE expression DO expression'
    t[0] = While(t[2], t[4])

# --------------------------------------------------------------
#  if then else

def p_expression_if_then_else(t):
    'expression : IF expression THEN expression ELSE expression'
    """
    Handles the parsing of if-then-else expressions.

    The function processes if-then-else constructs in the language. For an input of the form
    'IF expression THEN expression ELSE expression', it constructs an IfThenElse object with the 
    conditional expression, the expression to evaluate if the condition is true, and the expression 
    to evaluate if the condition is false, as arguments.

    Parameters:
    t: A list where each element represents a component of the syntax being parsed. The elements
       correspond to the 'IF expression THEN expression ELSE expression' structure.

    Sets:
    t[0]: An IfThenElse object, representing the if-then-else construct.
    """
    t[0] = IfThenElse(t[2], t[4], t[6])


def p_expression_if_then(t):
    'expression : IF expression THEN expression'
    """
    Handles the parsing of if-then expressions.

    The function processes if-then constructs in the language. For an input of the form
    'IF expression THEN expression', it constructs an IfThen object with the 
    conditional expression and the expression to evaluate if the condition is true, as arguments.

    Parameters:
    t: A list where each element represents a component of the syntax being parsed. The elements
       correspond to the 'IF expression THEN expression' structure.

    Sets:
    t[0]: An IfThen object, representing the if-then construct.
    """
    t[0] = IfThen(t[2], t[4])


# --------------------------------------------------------------
#  Lazy

def p_lazy_expr(t):
    'expression : LAZY expression'
    """
    Handles the parsing of 'LAZY expression' constructs in the language.

    The function processes constructs beginning with 'LAZY'. For an input of the form
    'LAZY expression', it constructs a Lazy object with the expression as an argument.

    Parameters:
    t: A list where each element represents a component of the syntax being parsed. The elements
       correspond to the 'LAZY expression' structure.

    Sets:
    t[0]: A Lazy object, representing the lazy evaluation construct.
    """
    t[0] = Lazy(t[2])  # Wrap the expression in a Lazy node.


def p_expression_braces(t):
    'expression : LCBRACE expression RCBRACE'
    """
    Handles the parsing of expressions within braces in the language.

    The function processes constructs that are enclosed within braces '{}'. For an input of the form
    '{expression}', it constructs a Lazy object with the expression as an argument.

    Parameters:
    t: A list where each element represents a component of the syntax being parsed. The elements
       correspond to the '{expression}' structure.

    Sets:
    t[0]: A Lazy object, representing the expression enclosed in braces.
    """
    t[0] = Lazy(t[2])  # Wrap the expression in a Lazy node.


def p_expression_lambda(t):
    'expression : LAMBDA LPAREN args RPAREN expression'
    """
    Handles the parsing of lambda expressions in the language.

    The function processes constructs of the form 'lambda (args) expression'. It creates a LambdaFunction 
    object with the arguments and the body expression as parameters.

    Parameters:
    t: A list where each element represents a component of the syntax being parsed. The elements
       correspond to the 'lambda (args) expression' structure.

    Sets:
    t[0]: A LambdaFunction object, representing the lambda expression.
    """
    t[0] = LambdaFunction(t[3], t[5])  # Create a new LambdaFunction node.


def p_expression_pipe(p):
    'expression : expression PIPE expression'
    """
    Handles the parsing of pipe operations in the language.

    The function processes constructs of the form 'expression | expression'. It creates a Pipe 
    object with the left and right expressions as parameters.

    Parameters:
    p: A list where each element represents a component of the syntax being parsed. The elements
       correspond to the 'expression | expression' structure.

    Sets:
    p[0]: A Pipe object, representing the pipe operation.
    """
    p[0] = Pipe(p[1], p[3])  # Create a new Pipe node.

def p_expression_func_call(t):
    'expression : function_call'
    """
    Handles the parsing of function calls in the language.

    The function processes constructs where a function is called. It sets the parsed function call
    as the expression's value.

    Parameters:
    t: A list where each element represents a component of the syntax being parsed. The elements
       correspond to the 'function_call' structure.

    Sets:
    t[0]: The parsed function call.
    """
    t[0] = t[1]  # Set the function call as the value of the expression.


def p_expression_term(t):
    'expression : term'
    """
    Processes the parsing of an individual term in the language.

    This function is used to parse individual terms, which are the simplest element 
    of an expression in the language. When a term is encountered, the value of the 
    term is set as the value of the expression.

    Parameters:
    t: A list where each element represents a component of the syntax being parsed. 
       The elements correspond to the 'term' structure.

    Sets:
    t[0]: The parsed term.
    """
    t[0] = t[1]  # Set the term as the value of the expression.


def p_error(t):
    """
    Handles syntax errors in the parsed code.

    This function is invoked when a syntax error is detected in the parsed code. 
    It can handle two types of errors: one where an unexpected token is encountered, 
    and one where the parser reaches the end of the file unexpectedly (EOF). For the 
    former, the function prints the line and position of the error, as well as the 
    offending token and its value.

    Parameters:
    t: A PLY token instance containing information about the token that resulted in 
       an error, or None if the error was due to reaching EOF unexpectedly.

    Returns:
    None
    """
    if t is None:  # If the error is due to EOF.
        print("Syntax error at EOF")
    else:  # If the error is due to an unexpected token.
        line = t.lexer.lineno  # Line number where the error occurred.
        position = t.lexpos  # Position in the file where the error occurred.
        token = t.type  # Type of the offending token.
        value = t.value  # Value of the offending token.
        print(f"Syntax error at line {line}, position {position}")
        print(f"Offending token: {token}")
        print(f"Offending value: {value}")



# -------------------------------------------------------------------
# Rule for function calls

def p_function_definition(t):
    'function_definition : FUNC IDENTIFIER LPAREN args RPAREN LCBRACE expression RCBRACE'
    """
    Parses function definition expressions.

    This function parses a function definition, which consists of the 'func' keyword,
    followed by an identifier (the function's name), parentheses enclosing arguments 
    (if any), and a block of code enclosed in curly braces. 

    After successful parsing, a FunctionDef object is constructed with the function name, 
    arguments, and the expression (body of the function) as arguments.

    Parameters:
    t: A PLY lex token instance. It is a tuple where t[2] is the function's name, t[4] is 
       the function's arguments, and t[7] is the body of the function (an expression).

    Returns:
    None
    """
    # FunctionDef is a hypothetical class to store function definitions. This will need to be 
    # defined elsewhere in your code. It takes the function name, the arguments, and the body 
    # of the function (an expression).
    t[0] = FunctionDef(t[2], t[4], t[7]) 

def p_definitions(t):
    '''
    definitions : function_definition
                | definitions function_definition
    '''
    # if there's only one function definition
    if len(t) == 2:
        t[0] = [t[1]]
    # if there's more than one function definition
    else:
        t[1].append(t[2])
        t[0] = t[1]


# def p_function_call(t):
    '''function_call : function_call LPAREN args RPAREN
                     | IDENTIFIER LPAREN args RPAREN
                     | IDENTIFIER'''
    """
    Parses function call expressions.

    This function parses a function call, which may be a simple function call with or without 
    arguments, or a chained function call with or without arguments. It constructs a FunctionCall 
    object with the function's name and arguments.

    Parameters:
    t: A PLY lex token instance. If it is a simple function call, t[1] is the function's name 
       and t[3] are the function's arguments. If it is a chained function call, t[1] is the 
       previous function call, and t[3] are the arguments of the next function in the chain.

    Returns:
    None
    """
    # # Check if there are arguments in the function call, if not set it to an empty list
    # args = [] if len(t) == 2 else t[3]
    # # Get the function name (it can be an identifier or another function call in case of function chaining)
    # function_name = t[1]
    # # FunctionCall and Identifier are hypothetical classes to store function calls and identifiers.
    # # These will need to be defined elsewhere in your code. FunctionCall takes the function name 
    # # (wrapped in an Identifier instance) and the arguments of the function.
    # t[0] = FunctionCall(Identifier(function_name), args) 


def p_function_call(t):
    '''function_call : function_call LPAREN args RPAREN
                     | IDENTIFIER LPAREN args RPAREN
                     | IDENTIFIER'''
    if len(t) == 4 and isinstance(t[2], FunctionCall):
        # Function call following another function call
        # Treat it as currying
        t[0] = FunctionCall(t[2], [t[4]])
    elif len(t) == 2:
        # Single function call
        args = []
        function_name = t[1]
        t[0] = FunctionCall(Identifier(function_name), args)
    else:
        # Function call with arguments
        args = t[3]
        function_name = t[1]
        t[0] = FunctionCall(Identifier(function_name), args)


def p_args(t):
    '''args : args COMMA arg
            | arg
            | '''
    """
    Parses arguments of a function or lambda.

    This function parses the arguments passed to a function call or a lambda function in the source code. 
    The arguments could be none, one, or many. It returns a list of arguments.

    Parameters:
    t: A PLY lex token instance. In the case of multiple arguments, t[1] is the list of arguments parsed 
       so far, and t[3] is the next argument. In the case of a single argument, t[1] is the argument. 

    Returns:
    None
    """
    if len(t) == 1:
        # No arguments are passed in the function call or lambda function
        t[0] = []
    elif len(t) == 2:
        # Only one argument is passed
        t[0] = [t[1]]
    else:
        # Multiple arguments are passed
        t[1].append(t[3])  # Append the next argument to the list of arguments
        t[0] = t[1]  # Set the list of arguments as the parsed result


def p_arg(t):
    '''arg : expression'''
    """
    Parses a single argument.

    This function is responsible for parsing a single argument in a function call or a lambda function 
    in the source code. It takes the expression as an argument and returns it.

    Parameters:
    t: A PLY lex token instance. t[1] is the parsed expression representing the argument.

    Returns:
    None
    """
    # The expression parsed is the argument
    t[0] = t[1]


# -------------------------------------------------------------------
# Initialize the parser
parser = yacc.yacc()
AST.inject_parsing(parser.parse)
# -------------------------------------------------------------------

def eval(s):
    """
    Evaluates a given string as a maryChain program.

    This function takes a string representation of a program written in the maryChain language,
    parses it into an abstract syntax tree (AST), and then evaluates it. If the result of the evaluation is a Lazy node,
    it is further evaluated to get the final result. 

    Parameters:
    s (str): The string to be evaluated as a maryChain program.

    Returns:
    result: The final result of evaluating the maryChain program. It can be a value of any data type supported in maryChain, 
    or None if there was a syntax error in the program.
    """
    # Parse the input string to an AST using the defined parser
    ast = parser.parse(s)

    # If there was a syntax error, print an error message and return None
    if ast is None:
        print("Syntax Error")
        return None

    # Otherwise, evaluate the AST using the evaluate method of the AST class
    result = AST.evaluate(ast)

    # If the result is a Lazy node, evaluate it to get the final result
    if isinstance(result, Lazy):
        result = result.evaluate()

    # Return the result of the evaluation
    return result
